
/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



package sfclasses;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * Filter configuration rule<br>
 * Objects of this class contain the complete information about one
 * filter rule. They are used to store the data for displaying the
 * active configuration of a firewall and to store the global configuration
 * data.
 * @version 1.0 03 Dec 1996
 * @author Roland E. Schmid
 */

public class Rule implements Persistent {

  /**
   * Initialize an empty rule object. This constructor is used by the
   * persistence methods.
   */
  public Rule() {
  }

  /**
   * Initialize the rule with data from a template. This is used for
   * autoconfiguration. 
   * @param oldrule Previously generated rule from same template or null.<br>
   * If oldrule is not null, the active, changed, and warn fields of the
   * new rule are initialized according to the old rule.
   * @param tID template ID
   * @param sID sub ID. More than one rule can be generated from a server
   * template. The sub ID is either 0 (for non-server templates) or
   * 2 * hierarchy level - 1 for server templates.
   */
  public Rule(Rule oldrule, int tID, int sID) {
    if (oldrule == null) {
      active = true;
      changed = false;
      warn = false;
    }
    else {
      active = oldrule.active;
      changed = false;
      warn = oldrule.changed | oldrule.warn;
    }
    if (tID < Templates.numberTotal) {
      templateID = tID;
      subID = sID;
      // fetch template data
      comment = Templates.comment[tID];
      priority = Templates.priority[tID];
      level_num = Templates.notification[tID];
      fw_src_idx = Templates.fw_src_idx[tID];
      fw_dst_idx = Templates.fw_dst_idx[tID];
      fw_src_cnt = 0;
      fw_dst_cnt = 0;
      timeout_type = SF_RULE_STATIC;
      fw_rc = Templates.fw_rc[tID];
      fw_flags = Templates.fw_flags[tID];
      ttl = Templates.ttl[tID];
      protocol = Templates.protocol[tID];
      fw_rip_idx = 0; // rip address templates not implemented
      fw_rip_cnt = 0;
    }
    else {
      System.out.println("Error: Template "+tID+" does not exist!");
      templateID = -1;
    }
  }

  /**
   * Bring up a dialog box for editing the rule.
   * @param parent Parent frame
   * @param mgDomain Manage domain object holding the global configuration data.
   * @param rp Calling frame that will be notified when the dialog closes
   * @param newRule If true the rule will be deleted from the rule vector
   * when the user selects cancel.
   */
  public void editDialog(Frame parent, ManageDomain mgDomain, Refreshable rp, boolean newRule) {
    RuleEditDialog ed = new RuleEditDialog(parent, this, mgDomain, rp, newRule);
    ed.setVisible(true);
  }

  /**
   * Return the text to be displayed for the rule in the rule panel.
   * @param column Number of the rule panel column.
   * @param firewall When displaying the active rules of a firewall, this
   * parameter contains the corresponding firewall, otherwise null.
   * @param mgDomain Manage domain object holding the global configuration data.
   * @return Array of strings for each line to display.
   */
  public String[] getString(int column, Host firewall, ManageDomain mgDomain) {
    stringCount = 0;
    switch (column) {
      case 0: // num column
        addString("Rule "); // add rule number later
        if (firewall != null) { // sfc show
          if (timeout_type == SF_RULE_STATIC)
            addString("static");
          else if (timeout_type == SF_TIMEOUT_VALID) {
            addString("dynamic");
            String tstr = timeout.toString();
            addString(tstr.substring(tstr.indexOf(':')-2,tstr.indexOf(':')+6));
          }
          else 
            addString("invalid");
        }
        else { // global rules
          if (templateID > -1) {
            addString("autoconf");
            if (changed)
              addString("changed");
            if (warn)
              addString("warn");
          }
          if (priority)
            addString("priority");
        }
        break;
      case 1: // action column
        switch (fw_rc) {
          case FW_ACCEPT:
            addString("accept");
            break;
          case SF_RC_OBSERVE:
            addString("observe");
            break;
          case FW_BLOCK:
            addString("block");
            break;
          case SF_RC_TREJECT:
            addString("reject");
            if (firewall == null) // rule editor
              addString(" with tcp reset");
            else // sfc show: tcp reset and best are synonymous
              addString(" with tcp rst / best");
            break;
          case SF_RC_BEST:
            addString("reject");
            addString(" with best");
          case SF_RC_ECHO:
            addString("reject");
            addString(" with echo reply");
            break;
          case SF_RC_RNET:
            addString("reject");
            addString(" with net unreachable");
            break;
          case FW_REJECT:
            addString("reject");
            addString(" with host unreachable");
            break;
          case SF_RC_RPROTO:
            addString("reject");
            addString(" with protocol unreachable");
            break;
          case SF_RC_RPORT:
            addString("reject");
            addString(" with port unreachable");
            break;
          default:
            addString("unknown");
            break;
        }
        break;
      case 2: // protocol column
        // IP options, ttl (show them in the protocol column as they are not used very often)
        if ((fw_flags & SF_FW_CHECK_OPT) != 0) {
          addString("IP options");
          if ((fw_flags & SF_FW_OPT_RR) != 0)
            addString(" Record route");
          if ((fw_flags & SF_FW_OPT_TS) != 0)
            addString(" Timestamp");
          if ((fw_flags & SF_FW_OPT_SEC) != 0)
            addString(" Security");
          if ((fw_flags & SF_FW_OPT_LSR) != 0)
            addString(" Loose source route");
          if ((fw_flags & SF_FW_OPT_SSR) != 0)
            addString(" Strict source route");
          if ((fw_flags & SF_FW_OPT_SATID) != 0)
            addString(" Sat ID");
        }
        if ((fw_flags & SF_FW_CHECK_TTL) != 0) {
          StringBuffer sb = new StringBuffer("IP: TTL ");
          switch (fw_flags & SF_FW_TTL) {
            case SF_FW_TTL_EQUAL:
              sb.append("=");
              break;
            case SF_FW_TTL_LESS:
              sb.append("<");
              break;
            case SF_FW_TTL_GREATER:
              sb.append(">");
              break;
            case SF_FW_TTL_NOTEQUAL:
              sb.append("!=");
              break;
          }
          sb.append(ttl);
          addString(sb.toString());
        }
        // protocol
        int protflags = fw_flags & SF_FW_PROT;
        if (protflags == SF_FW_PROT_ALL) {
          addString("all protocols");
        }
        else if (protflags == SF_FW_PROT_RIP) {
          addString("RIP");
          print_addr(fw_rip_idx, fw_rip_cnt, firewall, mgDomain);
        }
        else if (protflags == SF_FW_CHECK_PROTOCOL) {
          switch (protocol) {
            case IPPROTO_ICMP:
              addString("ICMP");
              if ((fw_flags & SF_TYPE_MASK) != SF_ICMP_ALLTYPES) {
                if ((fw_flags & SF_ICMP_ECHOREPLY) != 0)
                  addString(" echo reply");
                if ((fw_flags & SF_ICMP_DEST_UNREACH) != 0)
                  addString(" dest. unreachable");
                if ((fw_flags & SF_ICMP_SOURCE_QUENCH) != 0)
                  addString(" source quench");
                if ((fw_flags & SF_ICMP_REDIRECT) != 0)
                  addString(" redirect");
                if ((fw_flags & SF_ICMP_ECHO) != 0)
                  addString(" echo request");
                if ((fw_flags & SF_ICMP_TIME_EXCEEDED) != 0)
                  addString(" time exceeded");
                if ((fw_flags & SF_ICMP_PARAMETERPROB) != 0)
                  addString(" parameter problem");
                if ((fw_flags & SF_ICMP_TIMESTAMP) != 0)
                  addString(" timestamp request");
                if ((fw_flags & SF_ICMP_TIMESTAMPREPLY) != 0)
                  addString(" timestamp reply");
                if ((fw_flags & SF_ICMP_INFO_REQUEST) != 0)
                  addString(" information request");
                if ((fw_flags & SF_ICMP_INFO_REPLY) != 0)
                  addString(" information reply");
                if ((fw_flags & SF_ICMP_ADDRESS) != 0)
                  addString(" address mask request");
                if ((fw_flags & SF_ICMP_ADDRESSREPLY) != 0)
                  addString(" address mask reply");
              }
              break;
            case IPPROTO_IGMP:
              addString("IGMP");
              if ((fw_flags & SF_TYPE_MASK) != SF_IGMP_ALLTYPES) {
                if ((fw_flags & SF_IGMP_HOST_MEMBERSHIP_QUERY) != 0)
                  addString(" membership query");
                if ((fw_flags & SF_IGMP_HOST_MEMBERSHIP_REPORT) != 0)
                  addString(" membership report");
                if ((fw_flags & SF_IGMP_HOST_LEAVE_MESSAGE) != 0)
                  addString(" leave message");
              }
              break;
            case IPPROTO_TCP:
              addString("TCP");
              break;
            default:
              if (protocol <= MAX_PROTOCOL)
                addString(protocols[protocol]);
              else
                addString("Prot. "+Integer.toString(protocol));
              break;
          } // switch (protocol)
        }
        if ((protflags == SF_FW_PROT_ALL ||
            (protflags == SF_FW_CHECK_PROTOCOL && protocol == IPPROTO_TCP)) &&
            fw_rc == FW_ACCEPT) {
          if ((fw_flags & SF_FTP_NO_ACTIVE) != 0 && (fw_flags & SF_FTP_NO_PASSIVE) != 0)
            addString("no FTP data conn.");
          else if ((fw_flags & SF_FTP_NO_ACTIVE) != 0)
            addString("no FTP active mode");
          else if ((fw_flags & SF_FTP_NO_PASSIVE) != 0)
            addString("no FTP passive mode");
        }
        break;
      case 3: // from column
        print_addr(fw_src_idx, fw_src_cnt, firewall, mgDomain);
        break;
      case 4: // to column
        print_addr(fw_dst_idx, fw_dst_cnt, firewall, mgDomain);
        break;
      case 5: // notification column
        if (level_num == 0)
          addString("None");
        else {
          Notification n;
          if (mgDomain != null) 
            n = Notification.getLevel(level_num, mgDomain.NLevels);
          else
            n = null;
          if (n != null && n.levelName.length() > 0)
            addString(n.levelName);
          else
            addString("Level "+Integer.toString(level_num));
        }
        protflags = fw_flags & SF_FW_PROT;
        if ((protflags == SF_FW_PROT_ALL ||
            (protflags == SF_FW_CHECK_PROTOCOL && protocol == IPPROTO_TCP)) &&
            fw_rc == FW_ACCEPT) 
          if ((fw_flags & SF_FTP_DATA_LOG) != 0)
            addString("FTP data conn.");
        break;
      case 6: // usage or valid-for column
        if (firewall == null) { // 'valid for firewalls' column
          int validcnt = 0;
          int j;
          for (j=0; j < validFor.length; j++)
            if (validFor[j])
              validcnt++;
          try {
            if (validcnt == validFor.length)
              addString("All");
            else if (validcnt == 0)
              addString("None");
            else if (validcnt <= (validFor.length / 2)) {
              for (j=0; j < validFor.length; j++)
                if (validFor[j])
                  addString(((Host)mgDomain.Firewalls.elementAt(j)).objectID);
            }
            else {
              for (j=0; j < validFor.length; j++)
                if (!validFor[j])
                  addString("Not "+((Host)mgDomain.Firewalls.elementAt(j)).objectID);
            }
          }
          catch (ArrayIndexOutOfBoundsException ignore) {}
        }
        if (usage > 0)
          addString("Count "+Integer.toString(usage));
        if (bytes > 8*GIGA)
          addString(Long.toString(bytes / GIGA)+" GB");
        else if (bytes > 8*MEGA)
          addString(Long.toString(bytes / MEGA)+" MB");
        else if (bytes > 8*KILO)
          addString(Long.toString(bytes / KILO)+" KB");
        else if (bytes > 0)
          addString(Long.toString(bytes)+" Bytes");
        break;
      default:
        addString("Unknown column!");
        break;
    } // switch
    return returnArray();
  }

  // print address information
  // For active rules, the address array is used
  // For the rule editor, address macros are used
  private void print_addr(int idx, int cnt, Host firewall, ManageDomain mgDomain) {
    StringBuffer sb;
    boolean tcp_or_udp = false;

    if (idx <= SF_FIRSTMACRO) {
      Macro m;
      if (mgDomain != null)
        m = Macro.getMacro(idx, mgDomain.Macros);
      else
        m = null;
      if (m != null && m.macroName.length() > 0) {
        addString(m.macroName);
        switch (m.macroType) {
          case Macro.MACRO_INSIDE:
            addString(" Inside");
            break;
          case Macro.MACRO_OUTSIDE:
            addString(" Outside");
            break;
          case Macro.MACRO_ADDRESSLIST:
            String str[] = m.addresses.printAddresses();
            for (int i=0; i < str.length; i++)
              addString(" "+str[i]);
            break;
          case Macro.MACRO_OWNADDR:
            addString(" Own addresses");
            break;
        }
        if (m.port != Macro.MACRO_FIRSTPORT || m.prend != Macro.MACRO_LASTPORT) {
          if (m.port == m.prend)
            addString(" Port "+m.port);
          else
            addString(" Ports "+m.port+".."+m.prend);
        }
      }
      else
        addString("Macro "+(Math.abs(idx)-SF_FIRSTMACRO_ABS));
      return;
    }

    if (firewall == null) {
      if (idx != 0)
        addString("ERROR!");
      return;
    }

    if (((fw_flags & SF_FW_PROT) == SF_FW_CHECK_PROTOCOL) &&
        (protocol == IPPROTO_TCP || protocol == IPPROTO_UDP))
      tcp_or_udp = true;

    if ((fw_flags & SF_FW_SRC_NEG) != 0)
      addString("all except:");

    for (int i = idx; i < (idx+cnt); i++) {
      sb = new StringBuffer();

      if (firewall.mask[i][0] != 0 || firewall.mask[i][1] != 0 ||
          firewall.mask[i][2] != 0 || firewall.mask[i][3] != 0) {
        addString(Utils.printIP(firewall.addr[i]));
        sb.append(" ");
        if (firewall.mask[i][0] != -1 || firewall.mask[i][1] != -1 ||
            firewall.mask[i][2] != -1 || firewall.mask[i][3] != -1)
          addString(" mask "+Utils.printIP(firewall.mask[i]));
      }

      if (tcp_or_udp && firewall.port[i] != 0 && firewall.prend[i] != -1) {
        sb.append("port "+Long.toString(Utils.unsign(firewall.port[i])));
        if (firewall.port[i] != firewall.prend[i])
          sb.append(".."+Long.toString(Utils.unsign(firewall.prend[i])));
        addString(sb.toString());
      }
    }

  }

  // String array construction
  private void addString(String s) {
    if (stringCount == stringArraySize)
      return;
    stringArray[stringCount] = s;
    stringCount++;
  }

  // String array construction
  private String[] returnArray() {
    String str[] = new String[stringCount];
    for (int i=0; i < stringCount; i++) {
      str[i] = stringArray[i];
      stringArray[i] = null;
    }
    stringCount = 0;
    return str;
  }

  // temporary variables for output generation
  private static final int stringArraySize = 100;
  private String stringArray[] = new String[stringArraySize];
  private int stringCount = 0;
  
  /**
   * Print the rule in configuration file format. This method is called when
   * generating a configuration file for a firewall.
   * @param ps Stream to write the output to
   * @param mgDomain Manage domain object holding the global configuration data.
   * @param host Firewall the configuration file is generated for
   */
  public void printRule(PrintWriter ps, ManageDomain mgDomain, Host host) {

    ps.println("# "+comment);
    if (!active) {
      ps.println("# rule inactive - skipped...");
      ps.println();
      return;
    }

    switch (fw_rc) {
      case FW_ACCEPT:
        ps.println("accept");
        break;
      case SF_RC_OBSERVE:
        ps.println("observe");
        break;
      case FW_BLOCK:
        ps.println("block");
        break;
      case SF_RC_TREJECT:
        ps.println("reject with tcp_reset");
        break;
      case SF_RC_BEST:
        ps.println("reject with best");
        break;
      case SF_RC_ECHO:
        ps.println("reject with echo_reply");
        break;
      case SF_RC_RNET:
        ps.println("reject with icmp_net_unreachable");
        break;
      case FW_REJECT:
        ps.println("reject with icmp_host_unreachable");
        break;
      case SF_RC_RPROTO:
        ps.println("reject with icmp_protocol_unreachable");
        break;
      case SF_RC_RPORT:
        ps.println("reject with icmp_port_unreachable");
        break;
      default:
        ps.println("# ERROR: unknown return code "+fw_rc);
        ps.println();
        return;
    }

    boolean comma = false;
    if ((fw_flags & SF_FW_CHECK_OPT) != 0) {
      ps.println("  options");
      if ((fw_flags & SF_FW_OPT_RR) != 0)
        comma = printComma(ps, "    ", comma, "record_route");
      if ((fw_flags & SF_FW_OPT_TS) != 0)
        comma = printComma(ps, "    ", comma, "timestamp");
      if ((fw_flags & SF_FW_OPT_SEC) != 0)
        comma = printComma(ps, "    ", comma, "security");
      if ((fw_flags & SF_FW_OPT_LSR) != 0)
        comma = printComma(ps, "    ", comma, "loose_source_route");
      if ((fw_flags & SF_FW_OPT_SSR) != 0)
        comma = printComma(ps, "    ", comma, "strict_source_route");
      if ((fw_flags & SF_FW_OPT_SATID) != 0)
        comma = printComma(ps, "    ", comma, "sat_id");    
    }
    if ((fw_flags & SF_FW_CHECK_TTL) != 0) {
      StringBuffer sb = new StringBuffer("ttl ");
      switch (fw_flags & SF_FW_TTL) {
        case SF_FW_TTL_EQUAL:
          sb.append("= ");
          break;
        case SF_FW_TTL_LESS:
          sb.append("< ");
          break;
        case SF_FW_TTL_GREATER:
          sb.append("> ");
          break;
        case SF_FW_TTL_NOTEQUAL:
          sb.append("!= ");
          break;
      }
      sb.append(ttl);
      comma = printComma(ps, "    ", comma, sb.toString());
    }

    comma = false;
    // protocol
    int protflags = fw_flags & SF_FW_PROT;
    if (protflags == SF_FW_PROT_ALL) {
      ps.println("  all");
    }
    else if (protflags == SF_FW_PROT_RIP) {
      ps.println("  rip ");
      printAddrConfig(ps, 4, mgDomain, host, fw_rip_idx);
    }
    else if (protflags == SF_FW_CHECK_PROTOCOL) {
      switch (protocol) {
        case IPPROTO_ICMP:
          ps.println("  icmp");
          if ((fw_flags & SF_TYPE_MASK) != SF_ICMP_ALLTYPES) {
            if ((fw_flags & SF_ICMP_ECHOREPLY) != 0)
              comma = printComma(ps, "    ", comma, "echo_reply");
            if ((fw_flags & SF_ICMP_DEST_UNREACH) != 0)
              comma = printComma(ps, "    ", comma, "destination_unreachable");
            if ((fw_flags & SF_ICMP_SOURCE_QUENCH) != 0)
              comma = printComma(ps, "    ", comma, "source_quench");
            if ((fw_flags & SF_ICMP_REDIRECT) != 0)
              comma = printComma(ps, "    ", comma, "redirect");
            if ((fw_flags & SF_ICMP_ECHO) != 0)
              comma = printComma(ps, "    ", comma, "echo_request");
            if ((fw_flags & SF_ICMP_TIME_EXCEEDED) != 0)
              comma = printComma(ps, "    ", comma, "time_exceeded");
            if ((fw_flags & SF_ICMP_PARAMETERPROB) != 0)
              comma = printComma(ps, "    ", comma, "parameter_problem");
            if ((fw_flags & SF_ICMP_TIMESTAMP) != 0)
              comma = printComma(ps, "    ", comma, "timestamp_request");
            if ((fw_flags & SF_ICMP_TIMESTAMPREPLY) != 0)
              comma = printComma(ps, "    ", comma, "timestamp_reply");
            if ((fw_flags & SF_ICMP_INFO_REQUEST) != 0)
              comma = printComma(ps, "    ", comma, "information_request");
            if ((fw_flags & SF_ICMP_INFO_REPLY) != 0)
              comma = printComma(ps, "    ", comma, "information_reply");
            if ((fw_flags & SF_ICMP_ADDRESS) != 0)
              comma = printComma(ps, "    ", comma, "address_mask_request");
            if ((fw_flags & SF_ICMP_ADDRESSREPLY) != 0)
              comma = printComma(ps, "    ", comma, "address_mask_reply");
          }
          break;
        case IPPROTO_IGMP:
          ps.println("  igmp");
          if ((fw_flags & SF_TYPE_MASK) != SF_IGMP_ALLTYPES) {
            if ((fw_flags & SF_IGMP_HOST_MEMBERSHIP_QUERY) != 0)
              comma = printComma(ps, "    ", comma, "host_membership_query");
            if ((fw_flags & SF_IGMP_HOST_MEMBERSHIP_REPORT) != 0)
              comma = printComma(ps, "    ", comma, "host_membership_report");
            if ((fw_flags & SF_IGMP_HOST_LEAVE_MESSAGE) != 0)
              comma = printComma(ps, "    ", comma, "host_leave_message");
          }
          break;
        case IPPROTO_TCP:
          ps.println("  tcp");
          break;
        case IPPROTO_UDP:
          ps.println("  udp");
          break;
        default:
          ps.print("  "+Integer.toString(protocol));
          if (protocol <= MAX_PROTOCOL)
            ps.print(" /* "+protocols[protocol]+" */");
          else
            ps.print(" /* Protocol "+Integer.toString(protocol)+" */");
          ps.println();
          break;
      } // switch (protocol)
    }
    if ((protflags == SF_FW_PROT_ALL ||
        (protflags == SF_FW_CHECK_PROTOCOL && protocol == IPPROTO_TCP)) &&
        fw_rc == FW_ACCEPT) {
      if ((fw_flags & SF_FTP_NO_ACTIVE) != 0 && (fw_flags & SF_FTP_NO_PASSIVE) != 0)
        comma = printComma(ps, "    ", comma, "ftp_none");
      else if ((fw_flags & SF_FTP_NO_ACTIVE) != 0)
        comma = printComma(ps, "    ", comma, "ftp_passive");
      else if ((fw_flags & SF_FTP_NO_PASSIVE) != 0)
        comma = printComma(ps, "    ", comma, "ftp_active");
      else 
        comma = printComma(ps, "    ", comma, "ftp_all");
      if ((fw_flags & SF_FTP_DATA_LOG) != 0)
        comma = printComma(ps, "    ", comma, "ftp_log_data_conn");
    }

    comma = false;
    if (fw_src_idx <= SF_FIRSTMACRO) {
      ps.print("  from ");
      printAddrConfig(ps, 4, mgDomain, host, fw_src_idx);
    }
    if (fw_dst_idx <= SF_FIRSTMACRO) {
      ps.print("  to ");
      printAddrConfig(ps, 4, mgDomain, host, fw_dst_idx);
    }
    ps.println("  notification_level "+level_num+";");

    ps.println();

  } //printRule

  private boolean printComma(PrintWriter ps, String s1, boolean comma, String s2) {
    ps.print(s1);
    if (comma)
      ps.print(",");
    ps.println(s2);
    return true;
  }

  private void printPorts(PrintWriter ps, Macro m) {
    if (m.port != Macro.MACRO_FIRSTPORT || m.prend != Macro.MACRO_LASTPORT) {
      ps.print("port "+m.port);
      if (m.port != m.prend)
        ps.print(" .. "+m.prend);
    }
  }

  private void printAddrConfig(PrintWriter ps, int indent, ManageDomain mgDomain, Host host, int idx) {
    String indentString = new String();
    for (int i=0; i < indent; i++)
      indentString += " ";
    if (idx <= SF_FIRSTMACRO) {
      Macro m = Macro.getMacro(idx, mgDomain.Macros);
      if (m == null) {
        ps.println(indentString+"/* ERROR: Macro "+(Math.abs(idx)-SF_FIRSTMACRO_ABS)+" not found! */");
        return;
      }
      ps.println("/* "+m.macroName+" */");
      switch (m.macroType) {
        case Macro.MACRO_INSIDE:
          ps.print(indentString+"inside ");
          printPorts(ps, m);
          ps.println();
          break;
        case Macro.MACRO_OUTSIDE:
          ps.print(indentString+"outside ");
          printPorts(ps, m);
          ps.println();
          break;
        case Macro.MACRO_PORTS_ONLY:
          ps.print(indentString);
          printPorts(ps, m);
          ps.println();
          break;
        case Macro.MACRO_ADDRESSLIST:
          m.addresses.printAddressesFormatted(ps, indent, m.port, m.prend);
          break;
        case Macro.MACRO_OWNADDR:
          host.HostAddresses.printAddressesFormatted(ps, indent, m.port, m.prend);
          break;
      }
    }
    else
      ps.println();
  }
  // persistence methods
  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    ps.writeInt("Rule=", templateID);
    ps.writeInt("subID=", subID);
    ps.writeBoolean("active=", active);
    ps.writeBoolean("changed=", changed);
    ps.writeBoolean("warn=", warn);
    ps.writeBoolean("priority=", priority);
    ps.writeInt("validForLength=", validFor.length);
    for (int i=0; i < validFor.length; i++)
      ps.writeBoolean("validFor=", validFor[i]);
    ps.writeString("comment=", comment);
    ps.writeInt("level_num=", level_num);
    ps.writeInt("timeout_type=", timeout_type);
    ps.writeInt("fw_rc=", fw_rc);
    ps.writeInt("fw_src_idx=", fw_src_idx);
    ps.writeInt("fw_src_cnt=", fw_src_cnt);
    ps.writeInt("fw_dst_idx=", fw_dst_idx);
    ps.writeInt("fw_dst_cnt=", fw_dst_cnt);
    ps.writeInt("fw_flags=", fw_flags);
    ps.writeInt("ttl=", ttl);
    ps.writeInt("protocol=", protocol);
    ps.writeInt("fw_rip_idx=", fw_rip_idx);
    ps.writeInt("fw_rip_cnt=", fw_rip_cnt);
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    templateID = ps.readInt("Rule=");
    subID = ps.readInt("subID=");
    active = ps.readBoolean("active=");
    changed = ps.readBoolean("changed=");
    warn = ps.readBoolean("warn=");
    priority = ps.readBoolean("priority=");
    int len = ps.readInt("validForLength=");
    validFor = new boolean[len];
    for (int i=0; i < len; i++)
      validFor[i] = ps.readBoolean("validFor=");
    comment = ps.readString("comment=");
    level_num = ps.readInt("level_num=");
    timeout_type = ps.readInt("timeout_type=");
    fw_rc = ps.readInt("fw_rc=");
    fw_src_idx = ps.readInt("fw_src_idx=");
    fw_src_cnt = ps.readInt("fw_src_cnt=");
    fw_dst_idx = ps.readInt("fw_dst_idx=");
    fw_dst_cnt = ps.readInt("fw_dst_cnt=");
    fw_flags = ps.readInt("fw_flags=");
    ttl = ps.readInt("ttl=");
    protocol = ps.readInt("protocol=");
    fw_rip_idx = ps.readInt("fw_rip_idx=");
    fw_rip_cnt = ps.readInt("fw_rip_cnt=");
  }


  // additional rule data for GUI configuration
  protected int templateID = -1;// ID for autoconfigured rules
  protected int subID = 0;	// sub ID for server rules
                                // 1 = level-1-rule
                                // 2 = from-outside-rule
                                // 3 = from-inside-rule
  protected boolean active = true;    // if active==false rule is ignored
  protected boolean changed = false;  // user has manually changed autoconfigured rule
  protected boolean warn = false;     // changed rule has been overwritten
  protected boolean priority = false; // priority rule
  protected boolean validFor[]; // rule is valid for specific firewalls,
                                // index from ManageDomain.Firewalls
  protected String comment = new String(); // comment

  // rule data from sf_global.h
  protected int fw_line;	// line number in config file

  protected int level_num = 0;	// number of notification level

  protected static final int SF_TIMEOUT_VALID          =  0;
  protected static final int SF_RULE_CONNECTION        = -1;
  protected static final int SF_RULE_STATIC            = -2;
  protected static final int SF_RULE_SPOOFING          = -3;
  protected static final int SF_RULE_OVERSIZED         = -4;
  protected int timeout_type = -2;

  protected Date timeout;	// timeout value for dynamic rules

  protected int fw_rc = FW_ACCEPT;		// return code on match
  protected static final int SF_RC_BEST                = 100;  
  protected static final int SF_RC_TREJECT             = 20; /* reject TCP or UDP style */
  protected static final int SF_RC_ECHO                = 21; /* generate ICMP echo request -- ICMP only */
  protected static final int SF_RC_OBSERVE             = 22; /* generate log entry, skip to next rule */
  protected static final int FW_BLOCK                  =  1;
  protected static final int FW_ACCEPT                 =  2;
  protected static final int FW_REJECT                 = -1;
  protected static final int SF_RC_RNET                = -2; /* reject packet with icmp_net_unreachable */
  protected static final int SF_RC_RPROTO              = -3; /* reject packet with icmp_protocol_unreachable */
  protected static final int SF_RC_RPORT               = -4; /* reject packet with icmp_port_unreachable */

  /**
   * Address information<br>
   * positive: index to address array<br>
   * zero:     any addresses<br>
   * -1..-4:   special values for dynamic rules<br>
   * <=-100:   address macro number
   */  
  protected int fw_src_idx;     // index to address array
  protected int fw_src_cnt;
  /**
   * Address information<br>
   * positive: index to address array<br>
   * zero:     any addresses<br>
   * -1..-4:   special values for dynamic rules<br>
   * <=-100:   address macro number
   */  
  protected int fw_dst_idx;
  protected int fw_dst_cnt;
  // idx-values <= -100 are macro numbers (absolute value - 100)
  // idx -100 (macro 0) inside / outside (for server access)   
  // idx -101 (macro 1) == inside
  // idx -102 (macro 2) == outside
  protected static final int SF_SOURCEHOST             = -1;
  protected static final int SF_SOURCENET              = -2;
  protected static final int SF_DESTHOST               = -3;
  protected static final int SF_DESTNET                = -4;
  protected static final int SF_FIRSTMACRO             = -100;
  protected static final int SF_FIRSTMACRO_ABS         = Math.abs(SF_FIRSTMACRO);

  protected int fw_flags;       // flags
  // flags constants see below

  protected int ttl;		// time to live

  protected int protocol;	// protocol
  protected static final int IPPROTO_ICMP = 1;
  protected static final int IPPROTO_IGMP = 2;
  protected static final int IPPROTO_TCP = 6;
  protected static final int IPPROTO_UDP = 17;

  /**
   * Address information<br>
   * positive: index to address array<br>
   * zero:     any addresses<br>
   * -1..-4:   special values for dynamic rules<br>
   * <=-100:   address macro number
   */  
  protected int fw_rip_idx;	// RIP addresses index to address array
  protected int fw_rip_cnt;

  protected int usage;		// usage counter
  protected long bytes;		// byte counter
  protected static final long GIGA = 1024*1024*1024;
  protected static final long MEGA = 1024*1024;
  protected static final long KILO = 1024;

  /*
  *    Values for "flags" field:
  *       bit         meaning
  *        0           log packet (notification_level > 0)
  *        1           check ttl field
  *        2-3         ttl: 00 -> ttl == value
  *                         01 -> ttl < value
  *                         10 -> ttl > value
  *                         11 -> ttl != value
  *        4           1 == match "from" if none of the specified
  *                         addresses match (used for "outside")
  *        5           1 == match "to" if none of the specified
  *                         addresses match (used for "outside")
  *        6           option record_route
  *        7           option timestamp
  *        8           option security
  *        9           option loose_source_route
  *       10           option strict_source_route
  *       11           option sat_id
  *       12-15        protocol: 0000  all protocols
  *                              0001  rip
  *                              1111  current protocol (dynamic rule)
  *       16-31        type (protocol dependent)
  */

  protected static final int SF_FW_LOG            = 0x0001; /* log packet */

  protected static final int SF_FW_CHECK_TTL      = 0x0002; /* check time to live field */
  protected static final int SF_FW_TTL            = 0x000E; /* mask for SF_FW_TTL */
  protected static final int SF_FW_TTL_EQUAL      = 0x0002; /* ttl == value */
  protected static final int SF_FW_TTL_LESS       = 0x0006; /* ttl < value */
  protected static final int SF_FW_TTL_GREATER    = 0x000A; /* ttl > value */
  protected static final int SF_FW_TTL_NOTEQUAL   = 0x000E; /* ttl != value */

  protected static final int SF_FW_SRC_NEG        = 0x0010; /* from condition negated */
  protected static final int SF_FW_DST_NEG        = 0x0020; /* to condition negated */

  protected static final int SF_FW_CHECK_OPT      = 0x0FC0; /* check for options */
  protected static final int SF_FW_OPT_RR         = 0x0040; /* record route */
  protected static final int SF_FW_OPT_TS         = 0x0080; /* time stamp */
  protected static final int SF_FW_OPT_SEC        = 0x0100; /* security */
  protected static final int SF_FW_OPT_LSR        = 0x0200; /* loose source route */
  protected static final int SF_FW_OPT_SSR        = 0x0400; /* strict source route */
  protected static final int SF_FW_OPT_SATID      = 0x0800; /* sat id */

  protected static final int SF_FW_PROT           = 0xF000; /* mask to isolate protocol */
  protected static final int SF_FW_PROT_ALL       = 0x0000; /* matches all protocols */
  protected static final int SF_FW_PROT_RIP       = 0x1000; /* rip */
  protected static final int SF_FW_CHECK_PROTOCOL = 0xE000; /* compare protocol field */
  protected static final int SF_FW_PROT_CURRENT   = 0xF000; /* match current protocol in dynamic rule */

/* definitions for protocol dependent type bits: */
  protected static final int SF_TYPE_MASK           = 0xFFFF0000; /* mask to isolate protcol dependent bits */

  protected static final int SF_FTP_DATA_LOG        = 0x00010000; /* log data connections (TCP and ALL only) */
  protected static final int SF_FTP_NO_ACTIVE       = 0x00020000; /* don't allow FTP active mode */
  protected static final int SF_FTP_NO_PASSIVE      = 0x00040000; /* don't allow FTP passive mode */
  protected static final int SF_CHECK_SEQUENCE      = 0x00080000; /* check sequence numbers */

  protected static final int SF_ICMP_ALLTYPES       = 0x00000000;
  protected static final int SF_ICMP_ECHOREPLY      = 0x00010000; /* Echo Reply                   */
  protected static final int SF_ICMP_DEST_UNREACH   = 0x00020000; /* Destination Unreachable      */
  protected static final int SF_ICMP_SOURCE_QUENCH  = 0x00040000; /* Source Quench                */
  protected static final int SF_ICMP_REDIRECT       = 0x00080000; /* Redirect (change route)      */
  protected static final int SF_ICMP_ECHO           = 0x00100000; /* Echo Request                 */
  protected static final int SF_ICMP_TIME_EXCEEDED  = 0x00200000; /* Time Exceeded                */
  protected static final int SF_ICMP_PARAMETERPROB  = 0x00400000; /* Parameter Problem            */
  protected static final int SF_ICMP_TIMESTAMP      = 0x00800000; /* Timestamp Request            */
  protected static final int SF_ICMP_TIMESTAMPREPLY = 0x01000000; /* Timestamp Reply              */
  protected static final int SF_ICMP_INFO_REQUEST   = 0x02000000; /* Information Request          */
  protected static final int SF_ICMP_INFO_REPLY     = 0x04000000; /* Information Reply            */
  protected static final int SF_ICMP_ADDRESS        = 0x08000000; /* Address Mask Request         */
  protected static final int SF_ICMP_ADDRESSREPLY   = 0x10000000; /* Address Mask Reply           */

  protected static final int SF_IGMP_ALLTYPES               = 0x00000000;
  protected static final int SF_IGMP_HOST_MEMBERSHIP_QUERY  = 0x00010000; /* From RFC1112 */
  protected static final int SF_IGMP_HOST_MEMBERSHIP_REPORT = 0x00020000; /* Ditto */
  protected static final int SF_IGMP_HOST_LEAVE_MESSAGE     = 0x00040000; /* An extra BSD seems to send */

  protected static final int SF_RIP_ADDR_NEG        = 0x00010000; /* match rip-addr if none of the specified */
 				   /* addresses match (used for "outside") */
 
  protected static final int MAX_PROTOCOL = 103;
  protected static final String protocols[] =
  {"Prot. 0",			/*  0  Reserved */
  "ICMP",			/*  1  Internet Control Message */
  "IGMP",			/*  2  Internet Group Management */
  "GGP",			/*  3  Gateway-to-Gateway */
  "IP encap.",			/*  4  IP in IP (encapsulation) */
  "ST",				/*  5  Stream */
  "TCP",			/*  6  Transmission Control */
  "CBT",			/*  7  CBT */
  "EGP",			/*  8  Exterior Gateway Protocol */
  "IGP",			/*  9  Any private interior gateway */
  "BBN-RCC-MON",		/* 10  BBN-RCC-MON BBN RCC Monitoring */
  "NVP-II",			/* 11  Network Voice Protocol */
  "PUP",			/* 12  PUP */
  "ARGUS",			/* 13  ARGUS */
  "EMCON",			/* 14  EMCON */
  "XNET",			/* 15  Cross Net Debugger */
  "CHAOS",			/* 16  Chaos */
  "UDP",			/* 17  User Datagram */
  "MUX",			/* 18  Multiplexing */
  "DCN-MEAS",			/* 19  Measurement Subsystems */
  "HMP",			/* 20  Host Monitoring */
  "PRM",			/* 21  Packet Radio Measurement */
  "XNS-IDP",			/* 22  XEROX NS IDP */
  "TRUNK-1",			/* 23  Trunk-1 */
  "TRUNK-2",			/* 24  Trunk-2 */
  "LEAF-1",			/* 25  Leaf-1 */
  "LEAF-2",			/* 26  Leaf-2 */
  "RDP",			/* 27  Reliable Data Protocol */
  "IRTP",			/* 28  Internet Reliable Transaction */
  "ISO-TP4",			/* 29  ISO Transport Protocol Class 4 */
  "NETBLT",			/* 30  Bulk Data Transfer Protocol */
  "MFE-NSP",			/* 31  MFE Network Services Protocol */
  "MERIT-INP",			/* 32  MERIT Internodal Protocol */
  "SEP",			/* 33  Sequential Exchange Protocol */
  "3PC",			/* 34  Third Party Connect Protocol */
  "IDPR",			/* 35  Inter-Domain Policy Routing Protocol */
  "XTP",			/* 36  XTP */
  "DDP",			/* 37  Datagram Delivery Protocol */
  "IDPR-CMTP",			/* 38  IDPR Control Message Transport Proto */
  "TP++",			/* 39  TP++ Transport Protocol */
  "IL",				/* 40  IL Transport Protocol */
  "IPv6",			/* 41  IPv6 */
  "SDRP",			/* 42  Source Demand Routing Protocol */
  "IPv6-Route",			/* 43  Routing Header for IPv6 */
  "IPv6-Frag",			/* 44  Fragment Header for IPv6 */
  "IDRP",			/* 45  Inter-Domain Routing Protocol */
  "RSVP",			/* 46  Reservation Protocol */
  "GRE",			/* 47  General Routing Encapsulation */
  "MHRP",			/* 48  Mobile Host Routing Protocol */
  "BNA",			/* 49  BNA */
  "ESP",			/* 50  Encap Security Payload for IPv6 */
  "AH",				/* 51  Authentication Header for IPv6 */
  "I-NLSP",			/* 52  Integrated Net Layer Security */
  "SWIPE",			/* 53  IP with Encryption */
  "NARP",			/* 54  NBMA Next Hop Resolution Protocol */
  "MOBILE",			/* 55  IP Mobility */
  "TLSP",			/* 56  Transport Layer Security Protocol */
  "SKIP",			/* 57  SKIP */
  "IPv6-ICMP",			/* 58  ICMP for IPv6 */
  "IPv6-NoNxt",			/* 59  No Next Header for IPv6 */
  "IPv6-Opts",			/* 60  Destination Options for IPv6 */
  "Prot. 61",			/* 61  Any host internal protocol */
  "CFTP",			/* 62  CFTP */
  "Prot. 63",			/* 63  Any local network */
  "SAT-EXPAK",			/* 64  SATNET and Backroom EXPAK */
  "KRYPTOLAN",			/* 65  Kryptolan */
  "RVD",			/* 66  MIT Remote Virtual Disk Protocol */
  "IPPC",			/* 67  Internet Pluribus Packet Core */
  "Prot. 68",			/* 68  Any distributed file system */
  "SAT-MON",			/* 69  SATNET Monitoring */
  "VISA",			/* 70  VISA Protocol */
  "IPCV",			/* 71  Internet Packet Core Utility */
  "CPNX",			/* 72  Computer Protocol Network Executive */
  "CPHB",			/* 73  Computer Protocol Heart Beat */
  "WSN",			/* 74  Wang Span Network */
  "PVP",			/* 75  Packet Video Protocol */
  "BR-SAT-MON",			/* 76  Backroom SATNET Monitoring */
  "SUN-ND",			/* 77  SUN ND PROTOCOL-Temporary */
  "WB-MON",			/* 78  WIDEBAND Monitoring */
  "WB-EXPAK",			/* 79  WIDEBAND EXPAK */
  "ISO-IP",			/* 80  ISO Internet Protocol */
  "VMTP",			/* 81  VMTP */
  "SECURE-VMTP",		/* 82  SECURE-VMTP */
  "VINES",			/* 83  VINES */
  "TTP",			/* 84  TTP */
  "NSFNET-IGP",			/* 85  NSFNET-IGP */
  "DGP",			/* 86  Dissimilar Gateway Protocol */
  "TCF",			/* 87  TCF */
  "EIGRP",			/* 88  EIGRP */
  "OSPFIGP",			/* 89  OSPFIGP */
  "Sprite-RPC",			/* 90  Sprite RPC Protocol */
  "LARP",			/* 91  Locus Address Resolution Protocol */
  "MTP",			/* 92  Multicast Transport Protocol */
  "AX.25",			/* 93  AX.25 Frames */
  "IPIP encap.",		/* 94  IP-within-IP Encapsulation Protocol */
  "MICP",			/* 95  Mobile Internetworking Control Pro. */
  "SCC-SP",			/* 96  Semaphore Communications Sec. Pro. */
  "ETHERIP",			/* 97  Ethernet-within-IP Encapsulation */
  "ENCAP",			/* 98  Encapsulation Header */
  "Prot. 99",			/* 99  Any private encryption scheme */
  "GMTP",			/* 100 GMTP */
  "IFMP",			/* 101 Ipsilon Flow Management Protocol */
  "PNNI",			/* 102 PNNI over IP */
  "PIM",			/* 103 Protocol Independent Multicast */
  "Prot. "			/* Template for 104-255 */
  };

  /* ICMP Types */
  protected static final String icmptype[] =
  {"echo reply",		/*  0 */
  "type 1",			/*  1 */
  "type 2",			/*  2 */
  "dest. unr.",			/*  3 */
  "src quench",			/*  4 */
  "redirect",			/*  5 */
  "alt. addr.",			/*  6 */
  "type 7",			/*  7 */
  "echo req.",			/*  8 */
  "router adv.",		/*  9 */
  "router sel.",		/* 10 */
  "time xceed",			/* 11 */
  "param probl",		/* 12 */
  "time req.",			/* 13 */
  "time reply",			/* 14 */
  "info req.",			/* 15 */
  "info reply",			/* 16 */
  "addr req.",			/* 17 */
  "addr reply",			/* 18 */
  "type 19",			/* 19 */
  "type 20",			/* 20 */
  "type 21",			/* 21 */
  "type 22",			/* 22 */
  "type 23",			/* 23 */
  "type 24",			/* 24 */
  "type 25",			/* 25 */
  "type 26",			/* 26 */
  "type 27",			/* 27 */
  "type 28",			/* 28 */
  "type 29",			/* 29 */
  "traceroute",			/* 30 */
  "data cerr",			/* 31 */
  "mobl redir",			/* 32 */
  "where r u",			/* 33 */
  "i am here",			/* 34 */
  "mobil req.",			/* 35 */
  "mobil repl",			/* 36 */
  "dom req.",			/* 37 */
  "dom reply",			/* 38 */
  "skip",			/* 39 */
  "photuris",			/* 40 */
  };

} // Rule

// dialog box for editing rules
class RuleEditDialog
	extends Dialog
	implements Refreshable, Serializable,
						 ActionListener, ItemListener, TextListener {

  public RuleEditDialog(Frame p, Rule r, ManageDomain md, Refreshable rp, boolean nr) {
    super(p, "Rule", true);

    parent = p;
    rule = r;
    mgDomain = md;
    callingObject = rp;
    newRule = nr; // if set: delete rule from vector on cancel
    gbl = new GridBagLayout();
    gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    // workaround for OS/2 Java bug: Lists must be >= 2 lines high
    int minListSize = 1;
    if (System.getProperty("os.name").equals("OS/2"))
      minListSize = 2;

    boolean useBorderPanel = true;

    // Action Panel
    // ============
    Panel ap;
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    GridBagLayout agbl = new GridBagLayout();
    GridBagConstraints agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    CheckboxGroup ag = new CheckboxGroup();
    acceptBox = new Checkbox("Accept", ag, rule.fw_rc == Rule.FW_ACCEPT);
		acceptBox.addItemListener(this);
    Utils.add_component(ap, acceptBox, agbl, agbc, 0, 0, 1, 1, 0, 0);
    blockBox = new Checkbox("Block", ag, rule.fw_rc == Rule.FW_BLOCK);
		blockBox.addItemListener(this);
    Utils.add_component(ap, blockBox, agbl, agbc, 0, 1, 1, 1, 0, 0);
    observeBox = new Checkbox("Observe", ag, rule.fw_rc == Rule.SF_RC_OBSERVE);
		observeBox.addItemListener(this);
    Utils.add_component(ap, observeBox, agbl, agbc, 0, 2, 1, 1, 0, 0);
    rejectBox = new Checkbox("Reject with", ag, rule.fw_rc < 0 || 
      rule.fw_rc == Rule.SF_RC_BEST || rule.fw_rc == Rule.SF_RC_TREJECT || 
      rule.fw_rc == Rule.SF_RC_ECHO);
		rejectBox.addItemListener(this);
    Utils.add_component(ap, rejectBox, agbl, agbc, 0, 3, 1, 1, 0, 0);
    ap.add(rejectBox);

    rejectChoice = new Choice();
		rejectChoice.addItemListener(this);
    for (int i=0; i < rejectNumber; i++)
      rejectChoice.add(rejectText[i]);
    if (rejectBox.getState()) 
      for (int i=0; i < rejectNumber; i++)
        if (rule.fw_rc == rejectRC[i])
          rejectChoice.select(rejectText[i]);
    Utils.add_component(ap, rejectChoice, agbl, agbc, 0, 4, 1, 1, 0, 0);

    Utils.add_component(this, ap, gbl, gbc, 0, 0, 4, 6, 25, 50);

    // Options Panel
    // =============
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    optionsBox = new Checkbox("IP Options");
		optionsBox.addItemListener(this);
    optionsBox.setState((rule.fw_flags & Rule.SF_FW_CHECK_OPT) != 0);
    Utils.add_component(ap, optionsBox, agbl, agbc, 0, 0, 3, 1, 0, 0);

    optionsList = new List(3, true);
    for (int i=0; i < optionsNumber; i++) 
      optionsList.add(optionsText[i]);
    if ((rule.fw_flags & Rule.SF_FW_CHECK_OPT) != 0)
      for (int i=0; i < optionsNumber; i++)
        if ((rule.fw_flags & optionsFlags[i]) != 0)
          optionsList.select(i);
    Utils.add_component(ap, optionsList, agbl, agbc, 0, 1, 3, 3, 100, 100);

    ttlBox = new Checkbox("TTL");
		ttlBox.addItemListener(this);
    ttlBox.setState((rule.fw_flags & Rule.SF_FW_CHECK_TTL) != 0);
    Utils.add_component(ap, ttlBox, agbl, agbc, 0, 4, 1, 1, 0, 0);

    ttlChoice = new Choice();
		ttlChoice.addItemListener(this);
    for (int i=0; i < ttlNumber; i++)
      ttlChoice.add(ttlText[i]);
    if ((rule.fw_flags & Rule.SF_FW_CHECK_TTL) != 0)
      for (int i=0; i < ttlNumber; i++)
        if ((rule.fw_flags & Rule.SF_FW_TTL) == ttlFlags[i])
          ttlChoice.select(ttlText[i]);
    Utils.add_component(ap, ttlChoice, agbl, agbc, 1, 4, 1, 1, 0, 0);

    ttlValue = new TextField(Integer.toString(rule.ttl), 3);
    Utils.add_component(ap, ttlValue, agbl, agbc, 2, 4, 1, 1, 0, 0);
    
    Utils.add_component(this, ap, gbl, gbc, 0, 6, 4, 6, 25, 50);

    // Protocol Panel
    // ==============
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    ag = new CheckboxGroup();

    pallBox = new Checkbox("All protocols", ag, 
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_PROT_ALL);
    pallBox.addItemListener(this);
		Utils.add_component(ap, pallBox, agbl, agbc, 0, 0, 2, 1, 0, 0);

    ptcpBox = new Checkbox("TCP", ag, 
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_CHECK_PROTOCOL &&
      rule.protocol == Rule.IPPROTO_TCP);
		ptcpBox.addItemListener(this);
    Utils.add_component(ap, ptcpBox, agbl, agbc, 0, 1, 2, 1, 0, 0);

    pudpBox = new Checkbox("UDP", ag,
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_CHECK_PROTOCOL &&
      rule.protocol == Rule.IPPROTO_UDP);
		pudpBox.addItemListener(this);
    Utils.add_component(ap, pudpBox, agbl, agbc, 0, 2, 2, 1, 0, 0);

    picmpBox = new Checkbox("ICMP", ag,
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_CHECK_PROTOCOL &&
      rule.protocol == Rule.IPPROTO_ICMP);
		picmpBox.addItemListener(this);
    Utils.add_component(ap, picmpBox, agbl, agbc, 0, 3, 2, 1, 0, 0);

    pigmpBox = new Checkbox("IGMP", ag,
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_CHECK_PROTOCOL &&
      rule.protocol == Rule.IPPROTO_IGMP);
		pigmpBox.addItemListener(this);
    Utils.add_component(ap, pigmpBox, agbl, agbc, 0, 4, 2, 1, 0, 0);

    pripBox = new Checkbox("RIP", ag,
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_PROT_RIP);
    Utils.add_component(ap, pripBox, agbl, agbc, 0, 5, 2, 1, 0, 0);
		pripBox.addItemListener(this);

    pnumBox = new Checkbox("Prot. #", ag,
      (rule.fw_flags & Rule.SF_FW_PROT) == Rule.SF_FW_CHECK_PROTOCOL &&
      rule.protocol != Rule.IPPROTO_TCP && rule.protocol != Rule.IPPROTO_UDP &&
      rule.protocol != Rule.IPPROTO_ICMP && rule.protocol != Rule.IPPROTO_IGMP);
		pnumBox.addItemListener(this);
    Utils.add_component(ap, pnumBox, agbl, agbc, 0, 6, 1, 1, 0, 0);
 
    pnumLabel = new Label("", Label.RIGHT);
    if (pnumBox.getState()) {
      pnumText = new TextField(Integer.toString(rule.protocol), 3);
      if (rule.protocol > 0 && rule.protocol <= Rule.MAX_PROTOCOL)
        pnumLabel.setText("("+Rule.protocols[rule.protocol]+")");
    }
    else
      pnumText = new TextField(3);
		pnumText.addTextListener(this);
    Utils.add_component(ap, pnumText, agbl, agbc, 1, 6, 1, 1, 0, 0);
    agbc.fill = GridBagConstraints.HORIZONTAL;
    Utils.add_component(ap, pnumLabel, agbl, agbc, 0, 7, 2, 1, 100, 0);
    agbc.fill = GridBagConstraints.NONE;
      
    Utils.add_component(this, ap, gbl, gbc, 4, 0, 4, 12, 25, 100);

    // FTP Panel
    // =========
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    FTPPanel = ap;
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    Utils.add_component(ap, new Label("FTP Settings"), agbl, agbc, 0, 0, 1, 1, 0, 0);

    ag = new CheckboxGroup();

    int actpassmask = Rule.SF_FTP_NO_ACTIVE | Rule.SF_FTP_NO_PASSIVE;
    FTPallBox = new Checkbox("Both", ag, 
      (rule.fw_flags & actpassmask) == 0);
		FTPallBox.addItemListener(this);
    Utils.add_component(ap, FTPallBox, agbl, agbc, 0, 1, 1, 1, 0, 0);

    FTPactBox = new Checkbox("Active mode", ag,
      (rule.fw_flags & actpassmask) == Rule.SF_FTP_NO_PASSIVE);
		FTPactBox.addItemListener(this);
    Utils.add_component(ap, FTPactBox, agbl, agbc, 0, 2, 1, 1, 0, 0);

    FTPpassBox = new Checkbox("Passive mode", ag,
      (rule.fw_flags & actpassmask) == Rule.SF_FTP_NO_ACTIVE);
		FTPpassBox.addItemListener(this);
    Utils.add_component(ap, FTPpassBox, agbl, agbc, 0, 3, 1, 1, 0, 0);

    FTPnoneBox = new Checkbox("None", ag,
      (rule.fw_flags & actpassmask) == actpassmask);
		FTPnoneBox.addItemListener(this);
    Utils.add_component(ap, FTPnoneBox, agbl, agbc, 0, 4, 1, 1, 0, 0);


    FTPlogBox = new Checkbox("Log data conn.",
													(rule.fw_flags & Rule.SF_FTP_DATA_LOG) != 0);
		FTPlogBox.addItemListener(this);
    Utils.add_component(ap, FTPlogBox, agbl, agbc, 0, 6, 1, 1, 0, 0);


    // ICMP Panel
    // ==========
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    ICMPPanel = ap;
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    Utils.add_component(ap, new Label("ICMP Types"), agbl, agbc, 0, 0, 1, 1, 0, 0);

    ICMPallBox = new Checkbox("All types");
		ICMPallBox.addItemListener(this);
    ICMPallBox.setState(picmpBox.getState() &&
      (rule.fw_flags & Rule.SF_TYPE_MASK) == Rule.SF_ICMP_ALLTYPES);
    Utils.add_component(ap, ICMPallBox, agbl, agbc, 0, 1, 1, 1, 0, 0);
 
    ICMPList = new List(10, true);
    for (int i=0; i < ICMPNumber; i++)
      ICMPList.add(ICMPText[i]);
    if (picmpBox.getState())
      for (int i=0; i < ICMPNumber; i++)
        if ((rule.fw_flags & ICMPFlags[i]) != 0)
          ICMPList.select(i);
    Utils.add_component(ap, ICMPList, agbl, agbc, 0, 2, 1, 10, 100, 100);

    // IGMP Panel
    // ==========
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    IGMPPanel = ap;
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    Utils.add_component(ap, new Label("IGMP Types"), agbl, agbc, 0, 0, 1, 1, 0, 0);

    IGMPallBox = new Checkbox("All types");
		IGMPallBox.addItemListener(this);
    IGMPallBox.setState(picmpBox.getState() &&
      (rule.fw_flags & Rule.SF_TYPE_MASK) == Rule.SF_IGMP_ALLTYPES);
    Utils.add_component(ap, IGMPallBox, agbl, agbc, 0, 1, 1, 1, 0, 0);
 
    IGMPList = new List(10, true);
    for (int i=0; i < IGMPNumber; i++)
      IGMPList.add(IGMPText[i]);
    if (picmpBox.getState())
      for (int i=0; i < IGMPNumber; i++)
        if ((rule.fw_flags & IGMPFlags[i]) != 0)
          IGMPList.select(i);
    Utils.add_component(ap, IGMPList, agbl, agbc, 0, 2, 1, 10, 100, 100);

    // RIP Panel
    // =========
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    RIPPanel = ap;
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    ripList = new List(minListSize, false);
		ripList.addItemListener(this);
    Macro.constructList(ripList, mgDomain.Macros, rule.fw_rip_idx);
    Utils.add_component(ap, new Label("RIP Addresses"), agbl, agbc, 0, 0, 1, 1, 0, 0);
    Utils.add_component(ap, ripList, agbl, agbc, 0, 1, 1, 1, 0, 0);
    
    macroButton2 = new Button("New address macro...");
		macroButton2.addActionListener(this);
    Utils.add_component(ap, macroButton2, agbl, agbc, 0, 2, 1, 1, 0, 0);


    // add protocol specific panel
    // ===========================
    validPanel = null;
    if ((pallBox.getState() || ptcpBox.getState()) && acceptBox.getState()) {
      Utils.add_component(this, FTPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
      validPanel = FTPPanel;
    }
    if (picmpBox.getState()) {
      Utils.add_component(this, ICMPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
      validPanel = ICMPPanel;
    }
    if (pigmpBox.getState()) {
      Utils.add_component(this, IGMPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
      validPanel = IGMPPanel;
    }
    if (pripBox.getState()) {
      Utils.add_component(this, RIPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
      validPanel = RIPPanel;
    }


    // Others Panel: From, To, Priority, Valid for
    // ===========================================
    if (useBorderPanel)
      ap = new BorderPanel(BorderPanel.OUT);
    else
      ap = new Panel();
    othersPanel = ap;
    agbl = new GridBagLayout();
    agbc = new GridBagConstraints();
    agbc.fill = GridBagConstraints.NONE;
    agbc.anchor = GridBagConstraints.NORTHWEST;
    agbc.insets = new Insets(3,3,3,3);
    ap.setLayout(agbl);

    fromList = new List(minListSize, false);
		fromList.addItemListener(this);
    Macro.constructList(fromList, mgDomain.Macros, rule.fw_src_idx);
    Utils.add_component(ap, new Label("From"), agbl, agbc, 0, 0, 1, 1, 0, 0);
    Utils.add_component(ap, fromList, agbl, agbc, 1, 0, 1, 1, 0, 0);

    toList = new List(minListSize, false);
		toList.addItemListener(this);
    Macro.constructList(toList, mgDomain.Macros, rule.fw_dst_idx);
    Utils.add_component(ap, new Label("To"), agbl, agbc, 0, 1, 1, 1, 0, 0);
    Utils.add_component(ap, toList, agbl, agbc, 1, 1, 1, 1, 0, 0);

    macroButton = new Button("New address macro...");
		macroButton.addActionListener(this);
    Utils.add_component(ap, macroButton, agbl, agbc, 0, 2, 2, 1, 0, 0);


    notificationList = new List(minListSize, false);
		notificationList.addItemListener(this);
    Notification.constructList(notificationList, mgDomain.NLevels, rule.level_num);
    Utils.add_component(ap, new Label("Notification"), agbl, agbc, 0, 4, 1, 1, 0, 0);
    Utils.add_component(ap, notificationList, agbl, agbc, 1, 4, 1, 1, 0, 0);

    notificationButton = new Button("New level...");
		notificationButton.addActionListener(this);
    Utils.add_component(ap, notificationButton, agbl, agbc, 0, 5, 2, 1, 0, 0);


    priorityBox = new Checkbox("Priority rule");
		priorityBox.addItemListener(this);
    priorityBox.setState(rule.priority);
    Utils.add_component(ap, priorityBox, agbl, agbc, 0, 7, 2, 1, 0, 0);

    
    firewallList = new List(3, true);
		firewallList.addItemListener(this);
    for (int i=0; i < mgDomain.Firewalls.size(); i++) {
      firewallList.add(((Host)mgDomain.Firewalls.elementAt(i)).objectID);
      if (rule.validFor[i])
        firewallList.select(i);
    }
    Utils.add_component(ap, new Label("Rule valid on:"), agbl, agbc, 0, 9, 2, 1, 0, 0);
    Utils.add_component(ap, firewallList, agbl, agbc, 0, 10, 2, 3, 100, 100);

    Utils.add_component(this, ap, gbl, gbc, 12, 0, 4, 12, 25, 100);

    // Description
    // ===========
    Utils.add_component(this, new Label("Description:"), gbl, gbc, 0, 12, 3, 1, 0, 0);
    descText = new TextField(rule.comment);
    Utils.add_component(this, descText, gbl, gbc, 3, 12, 13, 1, 100, 0);

    // Buttons
    // =======
    gbc.fill = GridBagConstraints.NONE;
    gbc.anchor = GridBagConstraints.WEST;
    Panel pa = new Panel();
    OKbutton = new Button("OK");
		OKbutton.addActionListener(this);
    pa.add(OKbutton);
    Cancelbutton = new Button("Cancel");
		Cancelbutton.addActionListener(this);
    pa.add(Cancelbutton);
    Utils.add_component(this, pa, gbl, gbc, 0, 13, 16, 1, 0, 0);
    gbc.fill = GridBagConstraints.BOTH;
    gbc.anchor = GridBagConstraints.CENTER;

    // Window size
    pack();
		addWindowListener(new RDAdapter());
  }

class RDAdapter
	extends WindowAdapter
	implements Serializable {
	public void WindowClosing(WindowEvent we) {
		if (isEnabled())
			dispose();
		return;
	}
} // class RDAdapter

  public void actionPerformed(ActionEvent ae) {
	Object source = ae.getSource();

		// Process OKbutton
    if (source == OKbutton) {
      // get the information back into the rule structure
      if (collectData())
        dispose();
      if (callingObject != null)
        callingObject.refresh();
      return;
    }

		// Process Cancelbutton
    else if (source == Cancelbutton) {
      if (newRule)
        mgDomain.Rules.removeElement(rule);
      dispose();
      if (callingObject != null)
        callingObject.refresh();
      return;
    }

		// Macro Buttons
    else if (source == macroButton || source == macroButton2) {
      Macro ln;
      try {
        ln = (Macro)mgDomain.Macros.lastElement();
      }
      catch (NoSuchElementException e) {
        ln = null;
      }
      Macro m = new Macro
        (ln == null ? Macro.firstDynamic : ln.macroNumber - 1);
      mgDomain.Macros.addElement(m);
      addingMacro = true;
      m.edit(this, parent, mgDomain, true);
      return;
    }

		// Process notificationButton
    else if (source == notificationButton) {
      Notification ln;
      try {
        ln = (Notification)mgDomain.NLevels.lastElement();
      }
      catch (NoSuchElementException e) {
        ln = null;
      }
      Notification not = new Notification
        (ln == null ? Notification.firstDynamic : ln.levelNumber + 1);
      mgDomain.NLevels.addElement(not);
      addingNoti = true;
      not.edit(this, parent, mgDomain, true);
      return;
    }
	} // actionPerformed



	public void itemStateChanged(ItemEvent ie) {
	Object source = ie.getSource();
	
		// Process acceptBox
    if (source == acceptBox) {
      if ((pallBox.getState() || ptcpBox.getState()) && validPanel == null) {
        Utils.add_component(this, FTPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
        validPanel = FTPPanel;
        pack();
        validate();
      }
      return;
    }

		// Process blockBox, rejectBox, observeBox
    else if (source == blockBox || source == rejectBox || source == observeBox) {
      if (validPanel == FTPPanel) {
        remove(validPanel);
        validPanel = null;
        pack();
        validate();
      }
      return;
    }

		// Process pallBox, ptcpBox
    else if (source == pallBox || source == ptcpBox) {
      if (validPanel != FTPPanel) {
        boolean ch = false;
        if (validPanel != null) {
          ch = true;
          remove(validPanel);
          validPanel = null;
        }
        if (acceptBox.getState()) {
          ch = true;
          Utils.add_component(this, FTPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
          validPanel = FTPPanel;
        }
        if (ch) {
          pack();
          validate();
        }
      }
      return;
    }    

		// Process picmpBox
    else if (source == picmpBox) {
     if (validPanel != ICMPPanel) {
        if (validPanel != null)
          remove(validPanel);
        Utils.add_component(this, ICMPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
        validPanel = ICMPPanel;
        pack();
        validate();
      }
      return;
    }

		// Process pigmpBox
    else if (source == pigmpBox) {
     if (validPanel != IGMPPanel) {
        if (validPanel != null)
          remove(validPanel);
        Utils.add_component(this, IGMPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
        validPanel = IGMPPanel;
        pack();
        validate();
      }
      return;
    }

		// Process pripBox
    else if (source == pripBox) {
     if (validPanel != RIPPanel) {
        if (validPanel != null)
          remove(validPanel);
        Utils.add_component(this, RIPPanel, gbl, gbc, 8, 0, 4, 12, 25, 100);
        validPanel = RIPPanel;
        pack();
        validate();
      }
      return;
    }

		// Process pudpBox, pnumBox
    else if (source == pudpBox || source == pnumBox) {
      if (validPanel != null) {
        remove(validPanel);
        validPanel = null;
        pack();
        validate();
      }
      return;
    }
	} // itemStateChanged
		


	public void textValueChanged(TextEvent te) {
	Object source = te.getSource();

		// Process pnumText
    if (source == pnumText) {
      try {
        int pn = Integer.parseInt(pnumText.getText().trim());
        if (pn > 0 && pn <= Rule.MAX_PROTOCOL)
          pnumLabel.setText("("+Rule.protocols[pn]+")");
        else
          pnumLabel.setText("");
      }
      catch (NumberFormatException e) {
        pnumLabel.setText("");
      }
      validate();
      return;
    }
	} // textValueChanged



  public void refresh() {
    if (addingNoti) {
      notificationList.removeAll();
      Notification.constructList(notificationList, mgDomain.NLevels, 0);
      notificationList.select(notificationList.getItemCount()-1);
      notificationList.makeVisible(notificationList.getItemCount()-1);
      addingNoti = false;
    }

    if (addingMacro) {
      Macro newMacro = (Macro)mgDomain.Macros.lastElement();
      toList.add(newMacro.macroName);
      fromList.add(newMacro.macroName);
      ripList.add(newMacro.macroName);
      addingMacro = false;
    }
    validate();
  } // refresh


  private boolean collectData() {
    Macro m;
    rule.fw_flags = 0;

    // return code
    if (acceptBox.getState())
      rule.fw_rc = Rule.FW_ACCEPT;
    else if (blockBox.getState())
      rule.fw_rc = Rule.FW_BLOCK;
    else if (observeBox.getState())
      rule.fw_rc = Rule.SF_RC_OBSERVE;
    else if (rejectBox.getState()) 
      rule.fw_rc = rejectRC[rejectChoice.getSelectedIndex()];

    // IP options
    if (optionsBox.getState()) {
      int selidx[] = optionsList.getSelectedIndexes();
      for (int i=0; i < selidx.length; i++)
        rule.fw_flags |= optionsFlags[selidx[i]];
    }
    if (ttlBox.getState()) {
      rule.fw_flags |= ttlFlags[ttlChoice.getSelectedIndex()];
      boolean ttlerr = false;
      try {
        rule.ttl = Integer.parseInt(ttlValue.getText().trim());
      }
      catch (NumberFormatException e) {
        ttlerr = true;
      }
      if (ttlerr || rule.ttl < 0 || rule.ttl > 255) {
        UserDialog.ErrorBox("Invalid TTL value!");
        return false;
      }
    }

    // protocol
    if (pallBox.getState())
      rule.fw_flags |= Rule.SF_FW_PROT_ALL;
    else if (ptcpBox.getState()) {
      rule.fw_flags |= Rule.SF_FW_CHECK_PROTOCOL;
      rule.protocol = Rule.IPPROTO_TCP;
    }
    else if (pudpBox.getState()) {
      rule.fw_flags |= Rule.SF_FW_CHECK_PROTOCOL;
      rule.protocol = Rule.IPPROTO_UDP;
    }
    else if (picmpBox.getState()) {
      rule.fw_flags |= Rule.SF_FW_CHECK_PROTOCOL;
      rule.protocol = Rule.IPPROTO_ICMP;
      if (!ICMPallBox.getState()) {
        int selidx[] = ICMPList.getSelectedIndexes();
        for (int i=0; i < selidx.length; i++)
          rule.fw_flags |= ICMPFlags[selidx[i]];
      }
    }
    else if (pigmpBox.getState()) {
      rule.fw_flags |= Rule.SF_FW_CHECK_PROTOCOL;
      rule.protocol = Rule.IPPROTO_IGMP;
      if (!IGMPallBox.getState()) {
        int selidx[] = IGMPList.getSelectedIndexes();
        for (int i=0; i < selidx.length; i++)
          rule.fw_flags |= IGMPFlags[selidx[i]];
      }
    }
    else if (pripBox.getState()) {
      rule.fw_flags |= Rule.SF_FW_PROT_RIP;
      m = Macro.getMacro(ripList.getSelectedItem(), mgDomain.Macros);
      if (m != null)
        rule.fw_rip_idx = m.macroNumber;
      else {
        rule.fw_rip_idx = 0;
        rule.fw_rip_cnt = 0;
      }
    }
    else if (pnumBox.getState()) {
      boolean proterr = false;
      rule.fw_flags |= Rule.SF_FW_CHECK_PROTOCOL;
      try {
        rule.protocol = Integer.parseInt(pnumText.getText().trim());
      }
      catch (NumberFormatException e) {
        proterr = true;
      }
      if (proterr || rule.protocol < 0 || rule.protocol > 255) {
        UserDialog.ErrorBox("Invalid protocol number!");
        return false;
      }
    }
    
    // FTP settings
    if ((pallBox.getState() || ptcpBox.getState()) && acceptBox.getState()) {
      if (FTPactBox.getState() || FTPnoneBox.getState())
        rule.fw_flags |= Rule.SF_FTP_NO_PASSIVE;
      if (FTPpassBox.getState() || FTPnoneBox.getState())
        rule.fw_flags |= Rule.SF_FTP_NO_ACTIVE;
      if (FTPlogBox.getState())
        rule.fw_flags |= Rule.SF_FTP_DATA_LOG;
    }

    // various
    m = Macro.getMacro(fromList.getSelectedItem(), mgDomain.Macros);
    if (m != null)
      rule.fw_src_idx = m.macroNumber;
    else {
      rule.fw_src_idx = 0;
      rule.fw_src_cnt = 0;
    }
    m = Macro.getMacro(toList.getSelectedItem(), mgDomain.Macros);
    if (m != null)
      rule.fw_dst_idx = m.macroNumber;
    else {
      rule.fw_dst_idx = 0;
      rule.fw_dst_cnt = 0;
    }
    Notification n = Notification.getLevel(
      notificationList.getSelectedItem(), mgDomain.NLevels);
    if (n != null)
      rule.level_num = n.levelNumber;
    else
      rule.level_num = 0;
    rule.priority = priorityBox.getState();
    int fwidx[] = firewallList.getSelectedIndexes();
    for (int i=0; i < rule.validFor.length; i++)
      rule.validFor[i] = false;
    for (int i=0; i < fwidx.length; i++)
      rule.validFor[fwidx[i]] = true;
    rule.comment = descText.getText().trim();

    if (rule.templateID != -1) {
      rule.changed = true;
      rule.warn = false;
    }
    return true;
  } // collectDtat


  private static final int rejectNumber = 7;
  private static final String rejectText[] = {"Best",
                                              "Host unreachable",
                                              "Net unreachable",
                                              "Protocol unreach.",
                                              "Port unreachable",
                                              "TCP reset",
                                              "Echo reply"};
  private static final int rejectRC[] = {Rule.SF_RC_BEST,
                                         Rule.FW_REJECT,
                                         Rule.SF_RC_RNET,
                                         Rule.SF_RC_RPROTO,
                                         Rule.SF_RC_RPORT,
                                         Rule.SF_RC_TREJECT,
                                         Rule.SF_RC_ECHO};

  private static final int optionsNumber = 6;
  private static final String optionsText[] = {"Record route",
                                               "Time stamp",
                                               "Security",
                                               "Loose SR",
                                               "Strict SR",
                                               "Sat ID"};
  private static final int optionsFlags[] = {Rule.SF_FW_OPT_RR,
                                             Rule.SF_FW_OPT_TS,
                                             Rule.SF_FW_OPT_SEC,
                                             Rule.SF_FW_OPT_LSR,
                                             Rule.SF_FW_OPT_SSR,
                                             Rule.SF_FW_OPT_SATID};

  private static final int ttlNumber = 4;
  private static final String ttlText[] = {"=", "<", ">", "!="};
  private static final int ttlFlags[] = {Rule.SF_FW_TTL_EQUAL,
                                         Rule.SF_FW_TTL_LESS,
                                         Rule.SF_FW_TTL_GREATER,
                                         Rule.SF_FW_TTL_NOTEQUAL};

  private static final String ICMPText[] = {"Echo request",
                                            "Echo reply",
                                            "Dest. unreach.",
                                            "Source quench",
                                            "Redirect",
                                            "Time exceeded",
                                            "Parameter prob.",
                                            "Timestamp req.",
                                            "Timestamp repl.",
                                            "Info request",
                                            "Info reply",
                                            "Addr mask req.",
                                            "Addr mask repl."};
  private static final int ICMPFlags[] = { Rule.SF_ICMP_ECHO,
                                           Rule.SF_ICMP_ECHOREPLY,
                                           Rule.SF_ICMP_DEST_UNREACH,
                                           Rule.SF_ICMP_SOURCE_QUENCH,
                                           Rule.SF_ICMP_REDIRECT,
                                           Rule.SF_ICMP_TIME_EXCEEDED,
                                           Rule.SF_ICMP_PARAMETERPROB,
                                           Rule.SF_ICMP_TIMESTAMP,
                                           Rule.SF_ICMP_TIMESTAMPREPLY,
                                           Rule.SF_ICMP_INFO_REQUEST,
                                           Rule.SF_ICMP_INFO_REPLY,
                                           Rule.SF_ICMP_ADDRESS,
                                           Rule.SF_ICMP_ADDRESSREPLY };
  private static final int ICMPNumber = ICMPText.length;

  private static final String IGMPText[] = { "Membership query",
                                             "Membership report",
                                             "Leave message" };
  private static final int IGMPFlags[] = { Rule.SF_IGMP_HOST_MEMBERSHIP_QUERY,
                                           Rule.SF_IGMP_HOST_MEMBERSHIP_REPORT,
                                           Rule.SF_IGMP_HOST_LEAVE_MESSAGE };
  private static final int IGMPNumber = IGMPText.length;

  private Frame parent;
  private Rule rule;
  private ManageDomain mgDomain;
  private Refreshable callingObject;
  private boolean newRule;
 
  GridBagLayout gbl;
  GridBagConstraints gbc;
  private Button OKbutton;
  private Button Cancelbutton;
  private Panel ICMPPanel;
  private Panel IGMPPanel;
  private Panel FTPPanel;
  private Panel RIPPanel;
  private Panel validPanel;
  private Panel othersPanel;
  private Checkbox acceptBox;
  private Checkbox blockBox;
  private Checkbox rejectBox;
  private Checkbox observeBox;
  private Choice rejectChoice;
  private Checkbox optionsBox;
  private List optionsList;
  private Checkbox ttlBox;
  private Choice ttlChoice;
  private TextField ttlValue;
  private Checkbox pallBox;
  private Checkbox ptcpBox;
  private Checkbox pudpBox;
  private Checkbox picmpBox;
  private Checkbox pigmpBox;
  private Checkbox pripBox;
  private Checkbox pnumBox;
  private TextField pnumText;
  private Checkbox FTPallBox;
  private Checkbox FTPactBox;
  private Checkbox FTPpassBox;
  private Checkbox FTPnoneBox;
  private Checkbox FTPlogBox;
  private Checkbox ICMPallBox;
  private List ICMPList; 
  private Checkbox IGMPallBox;
  private List IGMPList;
  private List fromList;
  private List toList;
  private List notificationList;
  private Button notificationButton;
  private Button macroButton, macroButton2;
  private List firewallList;
  private Checkbox priorityBox;
  private TextField descText;
  private List ripList;
  private Label pnumLabel;
  private boolean addingNoti = false;
  private boolean addingMacro = false;
}
